#!/usr/bin/perl -w
#
#######################################################
# WebBench
# Written by: Brandon Zehm <caspian@linuxfreak.com>
#
# License:
# This software may be freely distributed, used, and
# modified, provided that this disclaimer and my name
# are included in the code.
#
# Disclaimer:
# -- insert standard disclaimer here ;) --
# I am not responsible for anything you do with this
# software, or anything that happens from the use of
# this software, including but not limited to anything
# you, your lawyers, or anyone else can think of.
#
#######################################################
#
# This file contains all the functions for WebBench.
# Any of these functions may be called from code you
# write and put in $conf{'script'}.
#
#
################################################################################
##  -------------------
##  Available Functions
##  -------------------
##    
##    contact("www.server.com")
##        Calls connectTo until a connection is established, printing
##        an error message(s) if/when connections are refused, etc.
##    
##    getPage("/page/name.html")
##        Requests a page from the web server, returning a reference
##        to an array containing the web page returned by the server.
##        If basicAuth() or setCookie() have been used getPage will
##        alter its behavior accordingly.
##    
##    basicAuth($name, $pass)             
##        Tells getPage to use the supplied name and password for 
##        basic authentication (does NOT support ssl!)
##    
##    getCookie(\@webPage)
##        Returns a ($name, $value) array when passed a reference
##        to a web page (most likely returned by getPage).  Returns
##        empty values if no cookie was found. Does not currently 
##        support mutiple name/value pairs.
##    
##    setCookie($name, $value)
##        Tells getPage to send the cookie specified when requesting
##        a page.  Does not currently support mutiple name/value pairs.
##    
##    getDate()
##        Returns the current date as a string in a human readable format.
##    
##    disconnect()
##        Disconnects from the web server.
##    
##    
##    
##    
##  --------------------------  
##  Functions for Internal Use
##  --------------------------
##    
##    bindSocket($name)
##        Opens a unix socket (named SOCKET) to a socket filename of $name.
##    
##    openSocket($name)
##        Connects to a unix socket (named SOCKET) to a socket filename of $name.
##    
##    closeSocket()
##        Closes the unix socket handle named SOCKET.
##        
##    connectTo("www.server.com", 80)
##      INTERNAL USE ONLY (use contact() from above)
##        Connects a local handle SERVER to the web server/port 
##        specified.  After calling this you are ready to use getPage 
##        to download a page.
##    
##    sendReport($lastReport)
##      INTERNAL USE ONLY
##        Connects to the reports server and sends it current stats.  Designed
##        to be run by the virtual clients reporting statistics to the Report
##        Server.
##    
##    printReport()
##      INTERNAL USE ONLY
##        Used by the reports generator to print the current stats to the 
##        screen.
##    
##
################################################################################








######################################################################
## Function: getPage($page)
##           $page is a url, it should not contain http://
##           or www.blah.com, it shoul look like this:
##
##           /site/index.html?name=value&blah=test
##      
##      
##           Returns a reference to an array of lines retrieved  
##           from the web server (i.e. the web page).
##
######################################################################
sub getPage {
  ## Get incoming variables
  ( 
    $incoming{'page'}
  ) = @_;
  chomp $incoming{'page'};
  
  ## Generate Request
  $global{'tmp'} =  "GET $incoming{'page'} HTTP/1.0$CRLF" . 
                    "Accept: \*\/\*$CRLF" . 
                    "User-Agent: WebBench-$conf{'version'} (Written by Brandon Zehm)$CRLF";
  $global{'tmp'} .= "Cookie: $global{'setCookie'}$CRLF" if ($global{'setCookie'} ne "");
  $global{'tmp'} .= "Authorization: Basic $global{'basicAuth'}$CRLF" if ($global{'basicAuth'} ne "");
  $global{'tmp'} .= "$CRLF";
  
  ## Request Page
  print SERVER $global{'tmp'};
  
  ## Update Stats
  $global{'totalSent'} += length($global{'tmp'});
  $global{'totalPages'} += 1;
  
  ## Get page from server
  my @tmp = ();
  while (<SERVER>) { 
    push (@tmp, $_);
    
    ## Update Stats
    $global{'totalReceived'} += length($_);
  }
  
  ## Return page
  return(\@tmp);
}












######################################################################
##  Function: getCookie (\@page)
##            &@page is a reference to an array containing a web page.
##            Each entry should be a line of the web page (such as is
##            returned from getPage().)
##            
##  Output:   returns an array like this:
##            ($name, $value)
##            Where $name is the name of the cookie and $value is the
##            content of the cookie. Returns an empty name and value 
##            if there were no cookies found.
##  Example:  getCookie(\@page)
##
##
##  TODO:     Support multiple cookie name/value pairs.
######################################################################
sub getCookie {
  $ref = $_[0];
  
  foreach(@$ref) {
    $tmp = $_;
      if ($tmp =~ /^Set-Cookie/) {
        my @cookie;
        @cookie = split(": ", $tmp);
        @cookie = split("=",  $cookie[1]);
        @tmp = split(";", $cookie[1]);
        $cookie[1] = $tmp[0];
        print "$$ - Cookie Name: $cookie[0]\n" if ($conf{'debug'} > 2);
        print "$$ - Cookie Value: $cookie[1]\n" if ($conf{'debug'} > 2);
        
        return($cookie[0], $cookie[1]);
      }
  }
  return('','');
}











######################################################################
##  Function: setCookie ($name, $value)
##              $name:   cookie name
##              $value:  cookie value
##            
##  Output:   returns 1 after setting global variable containing
##            information for getPage.  
##            
##  Example:  setCookie ($name, $value); getPage("/site/secure/index.html");
######################################################################
sub setCookie {
  ( 
    $incoming{'name'},
    $incoming{'value'}
  ) = @_;
  return (0) if ($incoming{'name'} eq "" or $incoming{'value'} eq "" );
  
  $global{'setCookie'} = "$incoming{'name'}=$incoming{'value'}";
  
  return(1);
}












######################################################################
##  Function: basicAuth ($login, $pass)
##              $login:  login name
##              $pass:   password
##            
##  Output:   returns an encoded string for basic auth, and sets a 
##            global variable telling
##            
##  Example:  basicAuth ($login, $pass); getPage("/site/secure/index.html");
######################################################################
sub basicAuth {
  ( 
    $incoming{'login'},
    $incoming{'pass'}
  ) = @_;
  
  $global{'basicAuth'} = "$incoming{'login'}:$incoming{'pass'}";
  
  ## Base64 encode the login and password
  my $padding = (3 - length($global{'basicAuth'}) % 3) % 3;                     ## Set flag if binary data isn't divisible by 3
  $global{'basicAuth'} = substr(pack('u', $global{'basicAuth'}), 1);            ## Convert the binary to uuencoded text
  chop($global{'basicAuth'});
  $global{'basicAuth'} =~ tr|` -_|AA-Za-z0-9+/|;                                ## Translate from uuencode to base64
  $global{'basicAuth'} =~ s/.{$padding}$/'=' x $padding/e if $padding;          ## Fix the end padding if flag (from above) is set
  
  
  return($global{'basicAuth'});
}









######################################################################
## Function: contact($server)
##           Connects $server:80 to a global socket named
##           SERVER.  Prints an error message if connection
##           is refused.  Returns 1 after successfully connecting.
######################################################################
sub contact {
  ## Get incoming variables
  ( 
    $incoming{'server'}
  ) = @_;
  
  until (connectTo($incoming{'server'}, 80)) {
    $date = getDate();
    print "$$ - $date Error connecting to server!\n";
  };
  return(1);
}






######################################################################
## Function: connectTo($server, $port)
##           Assumes port is 80 if $port is blank.
##           Connects $server:$post to a global socket named
##           SERVER
##           Returns 1 on success 0 on failure.
######################################################################
sub connectTo {
  ## Get incoming variables
  ( 
    $incoming{'server'},
    $incoming{'port'}
  ) = @_;
  
  print "$$ - connectTo() - sub entry\n" if ($conf{'debug'} > 5);
  my $return = 1;
  
  ## Check incoming variables
  $incoming{'port'} = 80 if ($incoming{'port'} eq "");
  print "$$ - ERROR: connectTo() Incoming \$server variable was empty.\n" if ($incoming{'server'} eq "");
  
  ## Open a IP socket in stream mode with tcp protocol. 
  print "$$ - connectTo() - requesting a streaming tcp/ip socket from the system\n" if ($conf{'debug'} > 5);
  socket(SERVER, PF_INET, SOCK_STREAM, getprotobyname('tcp')) || ($return = 0);
  print "$$ - ERROR:  Problem opening a tcp/ip socket with the system.\n" if ($return == 0);
  
  ## Create the data structure $dest by calling sockaddr_in(port, 32bit IP)
  print "$$ - connectTo() - creating data structure from server name and port\n" if ($conf{'debug'} > 5);
  $dest = sockaddr_in ($incoming{'port'}, inet_aton($incoming{'server'})) || ($return = 0);
  print "$$ - ERROR: Calling sockaddr_in() returned an error\n" if ($return == 0);
  
  ## Connect our socket to SERVER
  print "$$ - connectTo() - connecting the socket to the server\n" if ($conf{'debug'} > 5);
  connect(SERVER, $dest) || ($return = 0);
  print "$$ - ERROR:  Connection attempt to [$incoming{'server'}:$incoming{'port'}] failed!\n" if ($return == 0);
  print "$$ - connectTo() - successfully connected to $incoming{'server'}:$incoming{'port'}\n" if (($conf{'debug'} > 5) && ($return));
  
  ## Force our output to flush immediatly after a print.
  print "$$ - connectTo() - setting non-buffering mode on the network connection\n" if (($conf{'debug'} > 5) && ($return));
  select(SERVER);
  $| = 1;
  select(STDOUT);
  
  ## Return success
  print "$$ - connectTo() - sub exit: returning [$return]\n" if ($conf{'debug'} > 5);
  return($return);
}











######################################################################
# SUB: disconnect()
# Closes the SERVER socket.
# Returns 1 on success 0 on failure.
######################################################################
sub disconnect {
  print "$$ - disconnect() - sub entry\n" if ($conf{'debug'} > 5);
  my $return = 1;
  if (!(close SERVER)) {              ## and drop the connection.
    print "$$ - ERROR:  There was an error disconnecting form the server\n";
    $return = 0;                      ## Return failure if we didn't disconnect correctly
  } 
  print "$$ - disconnect() - disconnected from the server successfully\n" if ($conf{'debug'} > 5);
  print "$$ - disconnect() - sub exit: returning [$return]\n" if (($conf{'debug'} > 5) && ($return));
  return($return);                    ## Return
}







######################################################################
# SUB: getDate()
# Returns the current date in human readable string format.
# 
######################################################################
sub getDate {
  my ($sec, $min, $hour, $mday, $mon, $year);
  ($sec,$min,$hour,$mday,$mon,$year,undef,undef,undef) = localtime();
  
  return("$mon/$mday/$year $hour:$min:$sec");
}







#########################################################
# SUB: bindSocket ($name)
# Returns 1 on success 0 on failure.
#
# Connects the file $name to the socket SOCKET.
#########################################################
sub bindSocket {
  my $socket = $_[0];
  print "$$ - bindSocket() - sub entry\n" if ($conf{'debug'} > 5);
  my $return = 1;
  
  ## Open a UNIX socket in stream mode with tcp protocol. 
  print "$$ - bindSocket() - requesting a streaming unix socket from the system\n" if ($conf{'debug'} > 5);
  socket(SOCKET, PF_UNIX, SOCK_STREAM, 0) || ($return = 0);
  if ($return == 0) { print "$$ - bindSocket() - Error: $!\n"; }
  else { print "$$ - bindSocket() - Successfully requested a streaming unix socket from the system\n" if ($conf{'debug'} > 5); }
  
  ## Connect our socket to SOCKET
  print "$$ - bindSocket() - connecting the socket to a filehandle\n" if ($conf{'debug'} > 5);
  bind(SOCKET, sockaddr_un ($socket)) || ($return = 0);
  if ($return == 0) { print "$$ - bindSocket() - Error: $!\n"; }
  else { print "$$ - bindSocket() - successfully connected to $socket\n" if (($conf{'debug'} > 5) && ($return)); }
  
  ## Force our output to flush immediatly after a print.
  print "$$ - bindSocket() - setting non-buffering mode on the network connection\n" if (($conf{'debug'} > 5) && ($return));
  select(SOCKET);
  $| = 1;
  select(STDOUT);
  
  ## Return success
  print "$$ - bindSocket() - sub exit: returning [$return]\n" if ($conf{'debug'} > 5);
  return($return);
}







#########################################################
# SUB: openSocket ($name)
# Returns 1 on success 0 on failure.
#
# Connects the file $name to the socket SOCKET.
#########################################################
sub openSocket {
  my $socket = $_[0];
  print "$$ - openSocket() - sub entry\n" if ($conf{'debug'} > 5);
  my $return = 1;
  
  ## Open a UNIX socket in stream mode with tcp protocol. 
  print "$$ - openSocket() - requesting a streaming unix socket from the system\n" if ($conf{'debug'} > 5);
  socket(SOCKET, PF_UNIX, SOCK_STREAM, 0) || ($return = 0);
  if ($return == 0) { print "$$ - openSocket() - Error: $!\n"; }
  else { print "$$ - openSocket() - Successfully requested a streaming unix socket from the system\n" if ($conf{'debug'} > 5); }
  
  ## Connect our socket to SOCKET
  print "$$ - openSocket() - connecting the socket to a filehandle\n" if ($conf{'debug'} > 5);
  connect(SOCKET, sockaddr_un ($socket)) || ($return = 0);
  if ($return == 0) { print "$$ - openSocket() - Error: $!\n"; }
  else { print "$$ - openSocket() - successfully connected to $socket\n" if (($conf{'debug'} > 5) && ($return)); }
  
  ## Force our output to flush immediatly after a print.
  print "$$ - openSocket() - setting non-buffering mode on the network connection\n" if (($conf{'debug'} > 5) && ($return));
  select(SOCKET);
  $| = 1;
  select(STDOUT);
  
  ## Return success
  print "$$ - openSocket() - sub exit: returning [$return]\n" if ($conf{'debug'} > 5);
  return($return);
}







#########################################################
# SUB: closeSocket ()
# Closes the SOCKET socket.
# Returns 1 on success 0 on failure.
#########################################################
sub closeSocket {
  print "$$ - closeSocket() - sub entry\n" if ($conf{'debug'} > 5);
  my $return = 1;
  if (!(close SOCKET)) {              ## and drop the connection.
    print "$$ - ERROR:  There was an error disconnecting form the server\n";
    $return = 0;                      ## Return failure if we didn't disconnect correctly
  } 
  print "$$ - closeSocket() - disconnected from the server successfully\n" if (($conf{'debug'} > 5) && ($return));
  print "$$ - closeSocket() - sub exit: returning [$return]\n" if ($conf{'debug'} > 5);
  return($return);                    ## Return
}






######################################################################
## Function: sendReport($lastReport)
##           Connects to the reports generator and sends it some
##           information and resets local counters.
##           If $lastReport is a 1 it means the client is about to 
##           exit, so tell the report generator that the client is 
##           dead.  Use 0 normally.
##           Returns 0 on failure 1 on success.
######################################################################
sub sendReport {
  $global{'tmp'} = $_[0];
  
  print "$$ - sendReport() - sub entry\n" if ($conf{'debug'} > 5);
  
  if ($global{'tmp'} eq "") {
    $global{'tmp'} = 0
  }
  
  ## Send current data to the server
  unless (openSocket($conf{'socketName'})) {
    print "$$ - sendReport():  sub exit: returning [0]\n";
    closeSocket();
    return(0);
  }
  print SOCKET "$global{'totalPages'}:$global{'totalSent'}:$global{'totalReceived'}:$global{'tmp'}";
  closeSocket();
  
  ## Reset some variables
  $global{'totalPages'} = 0;
  $global{'totalSent'} = 0;
  $global{'totalReceived'} = 0;
  
  ## Done
  print "$$ - sendReport() - sub exit: returning [1]\n" if ($conf{'debug'} > 5);
  return(1);
}






######################################################################
## Function: printReport()
##           Connects to the reports generator and sends it some
##           information and resets local counters.
##           Returns 0 on failure 1 on success.
######################################################################
sub printReport {
  print "$$ - printReport() - sub entry\n" if ($conf{'debug'} > 5);
  
  ## Generate some numbers
  $global{'lstPagesSec'} = ($global{'totalPages'} - $global{'lstPagesSec'});
  $global{'avgPagesSec'} = ($global{'totalPages'} / ((time() + 1) - $^T));

print qq { 


Total Pages:                 $global{'totalPages'}
Total Bytes Sent:            $global{'totalSent'}
Total Bytes Received:        $global{'totalReceived'}

Average Pages/Second:        $global{'avgPagesSec'}
Pages Since Last Report:     $global{'lstPagesSec'}

};
  
  $global{'lstPagesSec'} = $global{'totalPages'};
  
  
  ## Done
  print "$$ - printReport() - sub exit: returning [1]\n" if ($conf{'debug'} > 5);
  return(1);
}







## Reap child processes.. is this nessacary?
sub reaper {
  $waitedpid = wait;
  # loathe sysV: it makes us not only reinstate
  # the handler, but place it after the wait
  $SIG{CHLD} = \&reaper;
}


## Quit
sub quit {
  print "$_[0]\n" if ($_[0]);
  exit(0);
}


## Quit because of a received signal
sub sigQuit {
  print "$$ - Exiting, received signal: $_[0]\n";
  exit(0);
}

